import abc
import argparse
import collections
import itertools
import os
import shutil
import sys
import tempfile

from unassigner.parse import parse_fasta, write_fasta
from unassigner.align import VsearchAligner, HitExtender
from unassigner.alignment import AlignedRegion


class TrimmableSeqs:
    def __init__(self, recs):
        replicate_seqs = collections.defaultdict(list)
        self.descs = dict()
        for desc, seq in recs:
            seq_id = desc.split()[0]
            replicate_seqs[seq].append(seq_id)
            self.descs[seq_id] = desc

        self.seqs = dict()
        self.seq_ids = dict()
        for seq, seq_ids in replicate_seqs.items():
            rep_seq_id = seq_ids[0]
            self.seqs[rep_seq_id] = seq
            self.seq_ids[rep_seq_id] = seq_ids
        self.matches = dict()

    def all_matched(self):
        return set(self.matches) == set(self.seq_ids)

    def get_matched_offset0(self):
        for rep_seq_id, matchobj in self.matches.items():
            if matchobj.offset == 0:
                seq = self.seqs[rep_seq_id]
                yield rep_seq_id, seq

    def get_unmatched_recs(self):
        for rep_seq_id in self.seq_ids:
            if rep_seq_id not in self.matches:
                seq = self.seqs[rep_seq_id]
                yield rep_seq_id, seq

    def get_replicate_ids(self, rep_seq_id):
        return self.seq_ids[rep_seq_id]

    def get_desc(self, seq_id):
        return self.descs[seq_id]

    def get_replicate_recs(self, rep_seq_id):
        seq_ids = self.seq_ids[rep_seq_id]
        seq = self.seqs[rep_seq_id]
        for seq_id in seq_ids:
            yield seq_id, seq

    def register_match(self, rep_seq_id, matchobj):
        self.matches[rep_seq_id] = matchobj

    @classmethod
    def from_fasta(cls, f):
        recs = parse_fasta(f)
        return cls(recs)


PrimerMatch = collections.namedtuple(
    "PrimerMatch", ["start", "end", "offset", "message"]
)


class Matcher(abc.ABC):
    def __init__(self, queryset):
        self.queryset = queryset

    # TODO: parallelize this
    def find_in_seqs(self, seqs):
        recs = seqs.get_unmatched_recs()
        for seq_id, seq in recs:
            match = self.find_match(seq)
            if match is not None:
                seqs.register_match(seq_id, match)
                yield seq_id, match

    def find_match(self, seq):
        raise NotImplemented()


class CompleteMatcher(Matcher):
    def __init__(self, queryset, max_mismatch):
        super().__init__(queryset)
        self.max_mismatch = max_mismatch

        # To look for near matches in the reference sequence, we
        # generate the set of qeury sequences with 0, 1, or 2 errors
        # and look for these in the reference.  This is our
        # "mismatched queryset."
        possible_mismatches = range(max_mismatch + 1)
        self.mismatched_queryset = [
            self._mismatched_queries(n) for n in possible_mismatches
        ]

    def _mismatched_queries(self, n_mismatch):
        # The generator is provides a sequence for one-time use, but
        # we need to go through it multiple times.  This function
        # wraps the generator to provide a list.
        return list(self._iter_mismatched_queries(n_mismatch))

    def _iter_mismatched_queries(self, n_mismatch):
        # This algorithm is terrible unless the number of mismatches is very small
        assert n_mismatch in [0, 1, 2, 3]
        for query in self.queryset:
            idx_sets = itertools.combinations(range(len(query)), n_mismatch)
            for idx_set in idx_sets:
                # Replace base at each position with a literal "N", to match
                # ambiguous bases in the reference
                yield replace_with_n(query, idx_set)
                # Change to list because strings are immutable
                qchars = list(query)
                # Replace the base at each mismatch position with an
                # ambiguous base specifying all possibilities BUT the one
                # we see.
                for idx in idx_set:
                    qchars[idx] = AMBIGUOUS_BASES_COMPLEMENT[qchars[idx]]
                    # Expand to all possibilities for mismatching at this
                    # particular set of positions
                for query_with_mismatches in deambiguate(qchars):
                    yield query_with_mismatches

    def find_match(self, seq):
        for n_mismatches, queryset in enumerate(self.mismatched_queryset):
            for query in queryset:
                start_idx = seq.find(query)
                if start_idx > -1:
                    if n_mismatches == 0:
                        msg = "Exact"
                    elif n_mismatches == 1:
                        msg = "Complete, 1 mismatch"
                    else:
                        msg = "Complete, {0} mismatches".format(n_mismatches)
                    end_idx = start_idx + len(query)
                    # obs_primer = seq[start_idx:end_idx]
                    return PrimerMatch(start_idx, end_idx, 0, msg)


def replace_with_n(seq, idxs):
    chars = list(seq)
    for idx in idxs:
        chars[idx] = "N"
    return "".join(chars)


class PartialMatcher(Matcher):
    def __init__(self, queryset, min_length):
        super().__init__(queryset)
        self.min_length = min_length

        self.partial_queries = set()
        for query in self.queryset:
            for partial_query in partial_seqs(query, self.min_length):
                self.partial_queries.add(partial_query)

    def find_match(self, seq):
        for partial_query in self.partial_queries:
            if seq.startswith(partial_query):
                end_idx = len(partial_query)
                return PrimerMatch(0, end_idx, 0, "Partial")


class AlignmentMatcher(Matcher):
    def __init__(
        self, alignment_dir, min_pct_id=75, min_aligned_frac=0.6, cores=1, suffix="0"
    ):
        assert os.path.exists(alignment_dir)
        assert os.path.isdir(alignment_dir)
        self.alignment_dir = alignment_dir
        self.min_pct_id = min_pct_id
        self.min_aligned_frac = min_aligned_frac
        self.cores = cores
        self.suffix = suffix

    def _make_fp(self, filename):
        return os.path.join(self.alignment_dir, filename)

    def find_in_seqs(self, seqs):
        if seqs.all_matched():
            return

        # Create the file paths
        subject_fp = self._make_fp("subject_{0}.fa".format(self.suffix))
        query_fp = self._make_fp("query_{0}.fa".format(self.suffix))
        result_fp = self._make_fp("query_{0}.txt".format(self.suffix))

        # Search
        with open(subject_fp, "w") as f:
            write_fasta(f, seqs.get_matched_offset0())
        ba = VsearchAligner(subject_fp)
        search_args = {"min_id": round(self.min_pct_id / 100, 2), "top_hits_only": None}
        if self.cores > 0:
            search_args["threads"] = self.cores
        hits = ba.search(
            seqs.get_unmatched_recs(),
            input_fp=query_fp,
            output_fp=result_fp,
            **search_args
        )

        # Refine
        bext = HitExtender(seqs.get_unmatched_recs(), seqs.get_matched_offset0())
        for hit in hits:
            alignment = bext.extend_hit(hit)
            subject_match = seqs.matches[alignment.subject_id]
            aligned_region = AlignedRegion.from_subject(
                alignment, subject_match.start, subject_match.end
            )
            query_start_idx, query_end_idx = aligned_region.in_query()
            query_offset = aligned_region.query_offset()
            matchobj = PrimerMatch(
                query_start_idx, query_end_idx, query_offset, "Alignment"
            )
            yield alignment.query_id, matchobj


def partial_seqs(seq, min_length):
    max_start_idx = len(seq) - min_length + 1
    for start_idx in range(1, max_start_idx):
        yield seq[start_idx:]


def aligned_frac(hit):
    unaligned_left = min(hit["qstart"], hit["sstart"])
    unaligned_right = min(
        hit["qlen"] - hit["qend"],
        hit["slen"] - hit["send"],
    )
    unaligned = unaligned_left + unaligned_right
    aligned = hit["length"]
    return aligned / (aligned + unaligned)


class TrimraggedApp(object):
    def __init__(self, seqs, trim_right, writer, min_trimmed_length):
        self.seqs = seqs
        self.trim_right = trim_right
        self.writer = writer
        self.min_trimmed_length = min_trimmed_length
        self.matchers = []

    def run(self):
        for matcher in self.matchers:
            matches = matcher.find_in_seqs(self.seqs)
            for rep_seq_id, matchobj in matches:
                if matchobj is not None:
                    self.seqs.register_match(rep_seq_id, matchobj)
                    seq_ids = self.seqs.seq_ids[rep_seq_id]
                    seq = self.seqs.seqs[rep_seq_id]
                    if self.trim_right:
                        trimmed_seq = trim_right(seq, matchobj)
                    else:
                        trimmed_seq = trim_left(seq, matchobj)
                    for seq_id in seq_ids:
                        desc = self.seqs.get_desc(seq_id)
                        if len(trimmed_seq) >= self.min_trimmed_length:
                            self.writer.write_trimmed(desc, trimmed_seq)
                        self.writer.write_stats(seq_id, seq, matchobj)

        for rep_seq_id, seq in self.seqs.get_unmatched_recs():
            for seq_id in self.seqs.get_replicate_ids(rep_seq_id):
                desc = self.seqs.get_desc(seq_id)
                self.writer.write_stats(seq_id, seq, None)


class Writer(object):
    def __init__(self, trimmed_file, stats_file):
        self.trimmed_file = trimmed_file
        self.stats_file = stats_file

    def write_trimmed(self, desc, seq):
        self.trimmed_file.write(">{0}\n{1}\n".format(desc, seq))

    def write_stats(self, seq_id, seq, matchobj):
        if matchobj is None:
            self.stats_file.write("{0}\tUnmatched\tNA\tNA\tNA\t\n".format(seq_id))
        else:
            matched_seq = seq[matchobj.start : matchobj.end]
            self.stats_file.write(
                "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\n".format(
                    seq_id,
                    matchobj.message,
                    matchobj.start,
                    matchobj.end,
                    matchobj.offset,
                    matched_seq,
                )
            )


def trim_left(seq, matchobj):
    return seq[matchobj.end :]


def trim_right(seq, matchobj):
    return seq[: matchobj.start]


def trim_middle(seq, matchobj):
    return seq[matchobj.start : matchobj.end]


def main(argv=None):
    p = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    p.add_argument("query", help="Query sequence to search and trim")
    p.add_argument("--input_file", type=argparse.FileType("r"), default=sys.stdin)
    p.add_argument(
        "--trimmed_output_file", type=argparse.FileType("w"), default=sys.stdout
    )
    p.add_argument(
        "--stats_output_file", type=argparse.FileType("w"), default=sys.stderr
    )

    # Matching parameters
    p.add_argument(
        "--max_mismatch",
        type=int,
        default=0,
        help="Maximum number of mismatches in complete match",
    )
    p.add_argument(
        "--min_partial",
        type=int,
        default=0,
        help=(
            "Minimum length of partial sequence match. " "Skip partial matching if 0."
        ),
    )
    p.add_argument(
        "--min_pct_id",
        type=float,
        default=70.0,
        help="Minimum percent identity in alignment stage",
    )

    # Overall program behavior
    p.add_argument(
        "--min_trimmed_length",
        type=int,
        default=1,
        help="Minimum length of trimmed output sequences",
    )
    p.add_argument(
        "--alignment_stages",
        type=int,
        default=1,
        help="Number of pairwise alignment stages",
    )
    p.add_argument(
        "--alignment_dir",
        help=(
            "Directory for files in alignment stage.  If not provided, "
            "a temporary directory will be used."
        ),
    )
    p.add_argument(
        "--cores",
        type=int,
        default=1,
        help="Number of CPU cores to use in alignment stage",
    )
    p.add_argument(
        "--reverse_complement_query",
        action="store_true",
        help="Reverse complement the query seq before search",
    )
    p.add_argument(
        "--trim_right",
        action="store_true",
        help="Trim right side, rather than left side of sequences",
    )

    args = p.parse_args(argv)

    seqs = TrimmableSeqs.from_fasta(args.input_file)
    writer = Writer(args.trimmed_output_file, args.stats_output_file)
    app = TrimraggedApp(seqs, args.trim_right, writer, args.min_trimmed_length)

    queryset = deambiguate(args.query)
    if args.reverse_complement_query:
        queryset = [reverse_complement(q) for q in queryset]

    if args.alignment_dir is None:
        alignment_dir = tempfile.mkdtemp()
    else:
        alignment_dir = args.alignment_dir
        if os.path.exists(alignment_dir):
            if not os.path.isdir(alignment_dir):
                raise RuntimeError(
                    "{0} exists and is not a directory".format(alignment_dir)
                )
        else:
            os.mkdir(alignment_dir)

    app.matchers.append(CompleteMatcher(queryset, args.max_mismatch))
    # TODO: Partial matching does not work on right hand side
    if args.min_partial > 0:
        app.matchers.append(PartialMatcher(queryset, args.min_partial))
    for n in range(args.alignment_stages):
        app.matchers.append(
            AlignmentMatcher(
                alignment_dir,
                min_pct_id=args.min_pct_id,
                cores=args.cores,
                suffix=str(n),
            )
        )

    app.run()

    if args.alignment_dir is None:
        shutil.rmtree(alignment_dir)


AMBIGUOUS_BASES = {
    "T": "T",
    "C": "C",
    "A": "A",
    "G": "G",
    "R": "AG",
    "Y": "TC",
    "M": "CA",
    "K": "TG",
    "S": "CG",
    "W": "TA",
    "H": "TCA",
    "B": "TCG",
    "V": "CAG",
    "D": "TAG",
    "N": "TCAG",
}


# Ambiguous base codes for all bases EXCEPT the key
AMBIGUOUS_BASES_COMPLEMENT = {
    "T": "V",
    "C": "D",
    "A": "B",
    "G": "H",
}


def deambiguate(seq):
    nt_choices = [AMBIGUOUS_BASES[x] for x in seq]
    return ["".join(c) for c in itertools.product(*nt_choices)]


COMPLEMENT_BASES = {
    "T": "A",
    "C": "G",
    "A": "T",
    "G": "C",
}


def reverse_complement(seq):
    rc = [COMPLEMENT_BASES[x] for x in seq]
    rc.reverse()
    return "".join(rc)



import argparse
import datetime
import gzip
import logging
import os
import pkg_resources

from unassigner.algorithm import (
    UnassignAligner,
    FileAligner,
    UnassignerApp,
    VariableMismatchRate,
)
from unassigner.parse import parse_fasta, parse_species_names
from unassigner.prepare_strain_data import download_type_strain_data


__version__ = pkg_resources.get_distribution("unassigner").version


def main(argv=None):
    p = argparse.ArgumentParser()
    p.add_argument(
        "query_fasta", type=argparse.FileType("r"), help="Query sequences FASTA file"
    )
    p.add_argument(
        "--output_dir",
        help=(
            "Output directory (default: basename of query sequences FASTA "
            "file, plus '_unassigned')"
        ),
    )
    p.add_argument(
        "--type_strain_fasta",
        default="unassigner_species.fasta",
        help=(
            "Type strain sequences FASTA file (default: %(default)s). "
            "If the default file is not found, sequences are downloaded "
            "and re-formatted automatically."
        ),
    )
    p.add_argument(
        "--threshold",
        type=float,
        help=(
            "Sequence identity threshold for ruling out species-level "
            "compatibility. Default value is 0.975 for the standard "
            "algorithm and 0.991 for the soft threshold algorithm."
        ),
    )
    p.add_argument(
        "--ref_mismatch_positions",
        help=(
            "File of mismatch positions in reference database. The file may "
            "be compressed in gzip format."
        ),
    )
    p.add_argument(
        "--num_cpus",
        type=int,
        help=(
            "Number of CPUs to use during sequence aligment (default: "
            "use all the CPUs)"
        ),
    )
    p.add_argument(
        "--soft_threshold", action="store_true", help="Use soft threshold algorithm."
    )
    p.add_argument("--verbose", action="store_true", help="Activate verbose mode.")
    p.add_argument("--version", action="version", version=f"{__version__}")
    args = p.parse_args(argv)

    if args.threshold is None:
        if args.soft_threshold:
            min_id = 0.991
        else:
            min_id = 0.975
    else:
        min_id = args.threshold

    if args.verbose is True:
        logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.INFO)

    logging.info(f"Starting at {datetime.datetime.now()}")

    query_seqs = list(parse_fasta(args.query_fasta, trim_desc=True))

    if args.output_dir is None:
        output_dir = os.path.splitext(args.query_fasta.name)[0] + "_unassigned"
    else:
        output_dir = args.output_dir

    # Download type strain files if needed
    type_strain_fp_is_default = args.type_strain_fasta == p.get_default(
        "type_strain_fasta"
    )
    type_strain_fp_is_missing = not os.path.exists(args.type_strain_fasta)
    if type_strain_fp_is_default and type_strain_fp_is_missing:
        download_type_strain_data()

    with open(args.type_strain_fasta) as f:
        species_names = dict(parse_species_names(f))

    writer = OutputWriter(output_dir, species_names)

    alignment_query_fp = writer.output_fp("unassigner_query.fasta")
    alignment_output_fp = writer.output_fp("unassigner_query_hits.txt")
    if os.path.exists(alignment_output_fp):
        a = FileAligner(args.type_strain_fasta, alignment_output_fp)
    else:
        a = UnassignAligner(args.type_strain_fasta)
        a.species_input_fp = alignment_query_fp
        a.species_output_fp = alignment_output_fp
        a.num_cpus = args.num_cpus

    if args.ref_mismatch_positions:
        if args.ref_mismatch_positions.endswith(".gz"):
            mm_db_file = gzip.open(args.ref_mismatch_positions, "rt")
        else:
            mm_db_file = open(args.ref_mismatch_positions)
        VariableMismatchRate.load_database(mm_db_file)

    app = UnassignerApp(
        a, VariableMismatchRate, min_id=min_id, soft_threshold=args.soft_threshold
    )
    for query_id, query_results in app.unassign(query_seqs):
        writer.write_results(query_id, query_results)

    logging.info(f"Finished at {datetime.datetime.now()}")


class OutputWriter:
    standard_keys = [
        "typestrain_id",
        "region_mismatches",
        "region_positions",
        "probability_incompatible",
    ]

    def __init__(self, output_dir, species_names):
        self.species_names = species_names

        if not os.path.exists(output_dir):
            os.mkdir(output_dir)
        self.output_dir = output_dir

        standard_fp = self.output_fp("unassigner_output.tsv")
        self.standard_file = open(standard_fp, "w")
        standard_header = ["query_id", "species"] + self.standard_keys
        self._write_tsv_line(self.standard_file, standard_header)

        algorithm_fp = self.output_fp("algorithm_output.tsv")
        self.algorithm_file = open(algorithm_fp, "w")
        self.algorithm_keys = None

    def output_fp(self, filename):
        return os.path.join(self.output_dir, filename)

    @staticmethod
    def _write_tsv_line(fileobj, vals):
        vals = [str(val) for val in vals]
        line = "\t".join(vals)
        fileobj.write(line)
        fileobj.write("\n")

    def write_results(self, query_id, results):
        for result in results:
            species_name = self.species_names.get(result["typestrain_id"], "NA")

            # Set custom algorithm keys based on the first result
            if self.algorithm_keys is None:
                self.algorithm_keys = list(result.keys())
                algorithm_header = ["query_id", "species_name"] + self.algorithm_keys
                self._write_tsv_line(self.algorithm_file, algorithm_header)

            standard_vals = [query_id, species_name] + [
                result[k] for k in self.standard_keys
            ]
            self._write_tsv_line(self.standard_file, standard_vals)

            algorithm_vals = [query_id, species_name] + [
                result[k] for k in self.algorithm_keys
            ]
            self._write_tsv_line(self.algorithm_file, algorithm_vals)



import collections
import logging
import os
import shutil
import subprocess
import urllib.request

from unassigner.parse import parse_fasta, parse_desc, parse_greengenes_accessions

LTP_METADATA_COLS = [
    "accession",
    "start",
    "stop",
    "unknown",
    "fullname_ltp",
    "type_ltp",
    "hi_tax_ltp",
    "riskgroup_ltp",
    "url_lpsn_ltp",
    "tax_ltp",
    "rel_ltp",
    "NJ_support_pk4_ltp",
]
LTP_VERSION = "08_2023"
LTP_METADATA_URL = (
    f"https://imedea.uib-csic.es/mmg/ltp/wp-content/uploads/ltp/LTP_{LTP_VERSION}.csv"
)
LTP_SEQS_URL = f"https://imedea.uib-csic.es/mmg/ltp/wp-content/uploads/ltp/LTP_{LTP_VERSION}_blastdb.fasta"
GG_SEQS_URL = "ftp://greengenes.microbio.me/greengenes_release/gg_13_5/gg_13_5.fasta.gz"
GG_ACCESSIONS_URL = (
    "ftp://greengenes.microbio.me/greengenes_release/gg_13_5/gg_13_5_accessions.txt.gz"
)
SPECIES_FASTA_FP = "unassigner_species.fasta"
REFSEQS_FASTA_FP = "refseqs.fasta"
GG_DUPLICATE_FP = "gg_duplicate_ids.txt"


def clean(db_dir):
    fps = [
        url_fp(LTP_METADATA_URL),
        url_fp(LTP_SEQS_URL),
        SPECIES_FASTA_FP,
        gunzip_fp(url_fp(GG_SEQS_URL)),
        gunzip_fp(url_fp(GG_ACCESSIONS_URL)),
        REFSEQS_FASTA_FP,
        GG_DUPLICATE_FP,
    ]
    for fp in fps:
        fp_full = os.path.join(db_dir, fp)
        if os.path.exists(fp_full):
            os.remove(fp_full)


def url_fp(url):
    return url.split("/")[-1]


def gunzip_fp(fp):
    return fp[:-3]


def get_url(url, fp):
    logging.info("Downloading {0}".format(url))
    with urllib.request.urlopen(url) as resp, open(fp, "wb") as f:
        shutil.copyfileobj(resp, f)
    return fp


def process_ltp_seqs(input_fp, output_fp=SPECIES_FASTA_FP):
    if os.path.isdir(output_fp):
        output_fp = os.path.join(output_fp, SPECIES_FASTA_FP)
    accession_cts = collections.defaultdict(int)
    # Re-format FASTA file
    with open(input_fp) as f_in:
        seqs = parse_fasta(f_in)
        with open(output_fp, "w") as f_out:
            for desc, seq in seqs:
                accession, species_name = parse_desc(desc)
                if not accession or not species_name:
                    continue
                # Some accessions refer to genomes with more than one 16S gene
                # So accessions can be legitiamtely repeated with distinct gene sequences
                accession_times_previously_seen = accession_cts[accession]
                accession_cts[accession] += 1
                if accession_times_previously_seen > 0:
                    accession = "{0}_repeat{1}".format(
                        accession, accession_times_previously_seen
                    )
                f_out.write(">{0}\t{1}\n{2}\n".format(accession, species_name, seq))
    return output_fp


def process_greengenes_seqs(seqs_fp, accessions_fp, output_fp=REFSEQS_FASTA_FP):
    duplicates_fp = GG_DUPLICATE_FP
    if os.path.isdir(output_fp):
        duplicates_fp = os.path.join(output_fp, duplicates_fp)
        output_fp = os.path.join(output_fp, REFSEQS_FASTA_FP)

    # Extract table of accessions
    if accessions_fp.endswith(".gz"):
        subprocess.check_call(["gunzip", "-f", accessions_fp])
        accessions_fp = gunzip_fp(accessions_fp)

    # Load accessions
    gg_accessions = {}
    with open(accessions_fp) as f:
        for ggid, src, acc in parse_greengenes_accessions(f):
            gg_accessions[ggid] = (acc, src)

    # Extract FASTA file
    if seqs_fp.endswith(".gz"):
        subprocess.check_call(["gunzip", "-f", seqs_fp])
        seqs_fp = gunzip_fp(seqs_fp)

    # Remove duplicate reference seqs
    uniq_seqs = collections.defaultdict(list)
    with open(seqs_fp) as f:
        for ggid, seq in parse_fasta(f):
            uniq_seqs[seq].append(ggid)

    with open(duplicates_fp, "w") as dups:
        with open(output_fp, "w") as f:
            for seq, ggids in uniq_seqs.items():
                ggid = ggids[0]
                if len(ggids) > 1:
                    dups.write(" ".join(ggids))
                # Re-label seqs with accession numbers
                acc, src = gg_accessions[ggid]
                f.write(">%s %s %s\n%s\n" % (acc, src, ggid, seq))

    return output_fp



import collections
import math
import operator

import numpy
from scipy.stats import betabinom

from unassigner.alignment import AlignedRegion
from unassigner.align import VsearchAligner, HitExtender
from unassigner.parse import parse_fasta


class UnassignAligner(object):
    def __init__(self, species_fp):
        self.species_fp = species_fp
        self.species_input_fp = None
        self.species_output_fp = None
        self.num_cpus = None

    def search_species(self, query_seqs):
        b = VsearchAligner(self.species_fp)
        vsearch_args = {
            "min_id": 0.9,
            "maxaccepts": 0,
        }
        if self.num_cpus:
            vsearch_args["threads"] = self.num_cpus
        hits = b.search(
            query_seqs, self.species_input_fp, self.species_output_fp, **vsearch_args
        )

        with open(self.species_fp) as f:
            ref_seqs = list(parse_fasta(f, trim_desc=True))
        xt = HitExtender(query_seqs, ref_seqs)
        for hit in hits:
            yield xt.extend_hit(hit)


class FileAligner:
    def __init__(self, species_fp, output_fp):
        self.species_fp = species_fp
        self.output_fp = output_fp

    def search_species(self, seqs):
        with open(self.species_fp) as f:
            ref_seqs = list(parse_fasta(f, trim_desc=True))
        xt = HitExtender(seqs, ref_seqs)
        with open(self.output_fp) as of:
            hits = VsearchAligner._parse(of)
            for hit in hits:
                yield xt.extend_hit(hit)


class VariableMismatchRate:
    """Predict unobserved mismatches by estimating a mismatch rate"""

    db = collections.defaultdict(list)
    result_keys = [
        "typestrain_id",
        "region_mismatches",
        "region_positions",
        "probability_incompatible",
        "mu1",
        "num_references",
        "mu2",
        "nonregion_positions_in_subject",
        "max_nonregion_mismatches",
    ]
    null_result = dict((key, "NA") for key in result_keys)

    @classmethod
    def clear_database(cls):
        cls.db.clear()

    @classmethod
    def load_database(cls, f):
        for line in f:
            line = line.rstrip()
            toks = line.split("\t")
            typestrain_id = toks[0]
            # ref_seq_id = toks[1]
            mismatch_positions = [int(x) for x in toks[2:]]
            cls.db[typestrain_id].append(mismatch_positions)

    @classmethod
    def _get_mismatches(self, typestrain_id, start_idx, end_idx):
        refs = self.db[typestrain_id]
        for ref_positions in refs:
            mismatch_is_in_region = [
                (pos >= start_idx) & (pos <= end_idx) for pos in ref_positions
            ]
            region_mms = mismatch_is_in_region.count(True)
            nonregion_mms = mismatch_is_in_region.count(False)
            yield (region_mms, nonregion_mms)

    def __init__(self, alignment):
        self.alignment = alignment
        self.query_id = alignment.query_id

    def unassign_threshold(self, min_id=0.975, soft_threshold=False):
        # Use all the beta-binomial logic from ConstantMismatchRate,
        # just adjust alpha and beta based on reference
        # sequences. Here's how. Reparameterize beta as mu and v,
        # following wikipedia. Hold v constant. We are going to update
        # mu. In the constant rate algorithm, mu2 = mu1. In the
        # variable rate algorithm, we determine log(mu2 / mu1) by
        # averaging the observed values from the reference
        # sequences. To stabilize things, start with a list of
        # [0,0,0,0,0]. Then, for each reference sequence, compute mu2
        # and mu1, take the log, and append to the list. Average the
        # values in the list. Now use this as the new value of mu2 for
        # the query sequence.

        # Clip out the aligned region
        region = AlignedRegion.without_endgaps(self.alignment)
        region_alignment = region.trim_ends()
        region_positions = region_alignment.alignment_len
        region_matches = region_alignment.count_matches()
        region_mismatches = region_positions - region_matches
        region_subject_positions = region_alignment.subject_len

        # Calcuate alpha, beta, mu, and v in aligned region
        alpha1 = region_mismatches + 0.5
        beta1 = region_matches + 0.5
        v1 = alpha1 + beta1
        mu1 = alpha1 / v1

        # Compute number of positions outside aligned region
        nonregion_subject_positions = (
            self.alignment.subject_len - region_subject_positions
        )
        total_positions = region_positions + nonregion_subject_positions

        # Get mismatches from database
        typestrain_id = self.alignment.subject_id
        typestrain_start_idx, typestrain_end_idx = region.in_subject()
        reference_mismatches = self._get_mismatches(
            typestrain_id, typestrain_start_idx, typestrain_end_idx
        )

        # Get estimate for gamma = log(mu2 / mu1)
        # From reference alignments
        reference_logvals = [0, 0, 0, 0, 0]
        for region_mms, nonregion_mms in reference_mismatches:
            # mu = alpha / (alpha + beta)
            # alpha = mismatches + 0.5
            # beta = matches + 0.5
            # matches = len - mismatches
            # beta = len - mismatches + 0.5
            # mu = (mismatches + 0.5) / (len + 1)
            ref_mu1 = (region_mms + 0.5) / (region_subject_positions + 1)
            ref_mu2 = (nonregion_mms + 0.5) / (nonregion_subject_positions + 1)
            log_mu2_mu1 = math.log(ref_mu2 / ref_mu1)
            reference_logvals.append(log_mu2_mu1)
        # TODO: add weighting
        gamma = numpy.mean(reference_logvals)

        # Calculate mu2, get alpha2 and beta2
        # log(mu2 / mu1) = gamma
        # log(mu2) - log(mu1) = gamma
        # log(mu2) = log(mu1) + gamma
        # mu2 = exp(log(mu1) + gamma)
        mu2 = math.exp(math.log(mu1) + gamma)
        v2 = v1
        # alpha2, beta2
        # mu2 = alpha2 / v2
        # alpha2 = mu2 * v2
        alpha2 = mu2 * v2
        # v2 = alpha2 + beta2
        beta2 = v2 - alpha2

        # Maximum number of mismatches outside observed region
        species_mismatch_threshold = 1 - min_id
        max_total_mismatches = int(
            math.floor(species_mismatch_threshold * total_positions)
        )
        max_nonregion_mismatches = max_total_mismatches - region_mismatches

        # Compute probability
        if soft_threshold:
            threshold_fcn = soft_species_probability
        else:
            threshold_fcn = hard_species_probability
        prob_compatible = threshold_assignment_probability(
            region_mismatches,
            region_positions,
            nonregion_subject_positions,
            alpha2,
            beta2,
            100 * species_mismatch_threshold,
            threshold_fcn,
        )
        prob_incompatible = 1 - prob_compatible

        return {
            "typestrain_id": self.alignment.subject_id,
            "region_mismatches": region_mismatches,
            "region_positions": region_positions,
            "probability_incompatible": prob_incompatible,
            "mu1": mu1,
            "num_references": len(reference_logvals),
            "mu2": mu2,
            "nonregion_positions_in_subject": nonregion_subject_positions,
            "max_nonregion_mismatches": max_nonregion_mismatches,
        }


def pctdiff(obs_mismatches, obs_positions, unobs_mismatches, unobs_positions):
    total_mismatches = obs_mismatches + unobs_mismatches
    total_positions = obs_positions + unobs_positions
    return 100 * (total_mismatches / total_positions)


def soft_species_probability(d, d_half):
    return math.pow(2, -d / d_half)


def hard_species_probability(d, d_half):
    return float(d <= d_half)


def iter_threshold(
    obs_mismatches,
    obs_positions,
    unobs_positions,
    alpha,
    beta,
    d_half,
    threshold_fcn=soft_species_probability,
):
    for mm in range(unobs_positions + 1):
        p_mm = betabinom.pmf(mm, unobs_positions, alpha, beta)
        d = pctdiff(obs_mismatches, obs_positions, mm, unobs_positions)
        p_species = threshold_fcn(d, d_half)
        if p_species < 1e-10:
            break
        yield (mm, p_mm, d, p_species)


def threshold_assignment_probability(
    obs_mismatches,
    obs_positions,
    unobs_positions,
    alpha,
    beta,
    d_half,
    threshold_fcn=soft_species_probability,
):
    return sum(
        p_mm * p_species
        for _, p_mm, _, p_species in iter_threshold(
            obs_mismatches,
            obs_positions,
            unobs_positions,
            alpha,
            beta,
            d_half,
            threshold_fcn,
        )
    )


class UnassignerApp:
    def __init__(self, aligner, mm_rate, min_id=0.975, soft_threshold=False):
        self.aligner = aligner
        self.mm_rate = mm_rate
        self.alignment_min_percent_id = min_id
        self.soft_threshold = soft_threshold

    def unassign(self, query_seqs):
        query_seqs = list(query_seqs)
        query_ids = [seq_id for seq_id, seq in query_seqs]

        # Step 1.
        # Align query sequences to type strain sequences. Same for all
        # algorithms.
        alignments = self._align_query_to_type_strain(query_seqs)

        # Step 2.
        # For each query-type strain alignment, estimate distribution of
        # mismatches outside fragment. Different for constant vs. variable
        # mismatch algorithms.
        mm_rates = [self.mm_rate(a) for a in alignments]

        # Step 3.
        # For each query-type strain alignment, estimate unassignment
        # probability. Different for hard vs. soft threshold algorithms.
        results = []
        for r in mm_rates:
            res = r.unassign_threshold(
                min_id=self.alignment_min_percent_id,
                soft_threshold=self.soft_threshold,
            )
            results.append((r.query_id, res))

        # Step 4.
        # Group by query and yield results to caller. Same for all algorithms.
        results_by_query = collections.defaultdict(list)
        for query_id, res in results:
            results_by_query[query_id].append(res)
        for query_id in query_ids:
            query_results = results_by_query[query_id]
            if not query_results:
                query_results = [self.mm_rate.null_result]
            yield query_id, query_results

    def _align_query_to_type_strain(self, query_seqs):
        # We expect query_seqs to be a list
        query_ids = [seq_id for seq_id, seq in query_seqs]
        unsorted_alignments = self.aligner.search_species(query_seqs)

        alignments_by_query = collections.defaultdict(list)
        for a in unsorted_alignments:
            alignments_by_query[a.query_id].append(a)

        for query_id in query_ids:
            query_alignments = alignments_by_query[query_id]
            query_alignments = self._filter_alignments(query_alignments)
            for a in query_alignments:
                yield a

    def _filter_alignments(self, query_alignments):
        sorted_alignments = list(
            sorted(
                query_alignments, key=operator.attrgetter("percent_id"), reverse=True
            )
        )
        filtered_alignments = [
            a for a in sorted_alignments if a.percent_id > self.alignment_min_percent_id
        ]
        # Return one low-identity result if we have nothing better
        if sorted_alignments and not filtered_alignments:
            return sorted_alignments[:1]
        return filtered_alignments



def species_probability(self, species_alignment, refseq_alignments):
    """Compute probability of species attribution."""
    query_id = species_alignment.query_id
    species_id = species_alignment.subject_id
    start = species_alignment.start_pos
    end = species_alignment.end_pos
    for r in refseq_alignments:
        # refseq_id = r.subject_id
        a, b = r.count_matches(start, end)
        c, d = r.count_matches()
        yield query_id, species_id, a, b, r.subject_id, c, d



import abc
import os.path
import subprocess
import tempfile
from Bio import pairwise2

from unassigner.parse import write_fasta, parse_fasta
from unassigner.alignment import AlignedPair

BLAST_FMT = (
    "qseqid sseqid pident length mismatch gapopen "
    "qstart qend sstart send qlen slen qseq sseq"
)
BLAST_FIELDS = BLAST_FMT.split()
BLAST_FIELD_TYPES = [
    str,
    str,
    float,
    int,
    int,
    int,
    int,
    int,
    int,
    int,
    int,
    int,
    str,
    str,
]


class Aligner(abc.ABC):
    def __init__(self, ref_seqs_fp):
        self.ref_seqs_fp = ref_seqs_fp

    def search(self, seqs, input_fp=None, output_fp=None, **kwargs):
        if input_fp is None:
            infile = tempfile.NamedTemporaryFile(mode="w+t", encoding="utf-8")
            write_fasta(infile, seqs)
            infile.seek(0)
            input_fp = infile.name
        else:
            with open(input_fp, "w") as f:
                write_fasta(f, seqs)

        if output_fp is None:
            outfile = tempfile.NamedTemporaryFile()
            output_fp = outfile.name

        self._call(input_fp, self.ref_seqs_fp, output_fp, **kwargs)

        with open(output_fp) as f:
            for hit in self._parse(f):
                yield hit

    @classmethod
    def _parse(self, f, convert_types=True):
        """Parse a BLAST output file."""
        for line in f:
            line = line.strip()
            if line.startswith("#"):
                continue
            vals = line.split("\t")
            if convert_types:
                vals = [fn(v) for fn, v in zip(BLAST_FIELD_TYPES, vals)]
            yield dict(zip(BLAST_FIELDS, vals))


class VsearchAligner(Aligner):
    @property
    def ref_seqs_udb_fp(self):
        base_fp, _ = os.path.splitext(self.ref_seqs_fp)
        return base_fp + ".udb"

    def make_reference_udb(self):
        if os.path.exists(self.ref_seqs_udb_fp):
            return
        args = [
            "vsearch",
            "--makeudb_usearch",
            self.ref_seqs_fp,
            "--output",
            self.ref_seqs_udb_fp,
        ]
        return subprocess.check_call(args)

    @staticmethod
    def _index(fasta_fp):
        pass

    def _call(self, query_fp, database_fp, output_fp, min_id=0.5, **kwargs):
        """Call the VSEARCH program.

        --id is a required argument to run the program. We expose the
        argument here as min_id.
        """
        if not os.path.exists(self.ref_seqs_udb_fp):
            self.make_reference_udb()
        args = [
            "vsearch",
            "--usearch_global",
            query_fp,
            "--db",
            self.ref_seqs_udb_fp,
            "--iddef",
            "2",
            "--id",
            str(min_id),
            "--userout",
            output_fp,
            "--userfields",
            "query+target+id2+alnlen+mism+gaps+qilo+qihi+tilo+tihi+qs+ts+qrow+trow",
        ]
        for arg, val in kwargs.items():
            arg = "--" + arg
            if val is None:
                args.append(arg)
            else:
                args += [arg, str(val)]
        subprocess.check_call(args, stderr=subprocess.DEVNULL)


class BlastAligner(Aligner):
    @staticmethod
    def _index(fasta_fp):
        return subprocess.check_call(
            [
                "makeblastdb",
                "-dbtype",
                "nucl",
                "-in",
                fasta_fp,
            ],
            stdout=subprocess.DEVNULL,
        )

    def _call(self, query_fp, database_fp, output_fp, **kwargs):
        """Call the BLAST program."""
        args = [
            "blastn",
            "-evalue",
            "1e-5",
            "-outfmt",
            "6 " + BLAST_FMT,
        ]
        for arg, val in kwargs.items():
            arg = "-" + arg
            if val is None:
                args.append(arg)
            else:
                args += [arg, str(val)]
        args += [
            "-query",
            query_fp,
            "-db",
            database_fp,
            "-out",
            output_fp,
        ]
        subprocess.check_call(args)


class HitExtender:
    def __init__(self, query_seqs, ref_seqs):
        self.query_seqs = dict(query_seqs)
        self.ref_seqs = dict(ref_seqs)

    def extend_hit(self, hit):
        # Handle the simple case where the local alignment covers both
        # sequences completely
        if self._is_global(hit):
            return AlignedPair(
                (hit["qseqid"], hit["qseq"]), (hit["sseqid"], hit["sseq"])
            )

        # We are going to need some repair or realignment.
        qseq = self.query_seqs[hit["qseqid"]]
        assert len(qseq) == hit["qlen"]
        sseq = self.ref_seqs[hit["sseqid"]]
        assert len(sseq) == hit["slen"]

        if self._needs_realignment(hit):
            aligned_qseq, aligned_sseq = align_semiglobal(qseq, sseq)
            return AlignedPair(
                (hit["qseqid"], aligned_qseq), (hit["sseqid"], aligned_sseq)
            )

        qleft, sleft = self._add_endgaps_left(hit, qseq, sseq)
        qright, sright = self._add_endgaps_right(hit, qseq, sseq)
        aligned_qseq = qleft + hit["qseq"] + qright
        aligned_sseq = sleft + hit["sseq"] + sright
        return AlignedPair((hit["qseqid"], aligned_qseq), (hit["sseqid"], aligned_sseq))

    @staticmethod
    def _is_global(hit):
        return (
            (hit["qstart"] == 1)
            and (hit["sstart"] == 1)
            and (hit["qend"] == hit["qlen"])
            and (hit["send"] == hit["slen"])
        )

    @staticmethod
    def _needs_realignment(hit):
        more_to_the_left = (hit["qstart"] > 1) and (hit["sstart"] > 1)
        more_to_the_right = (hit["qend"] < hit["qlen"]) and (hit["send"] < hit["slen"])
        return more_to_the_left or more_to_the_right

    @staticmethod
    def _add_endgaps_left(hit, qseq, sseq):
        # No repair needed
        if (hit["qstart"] == 1) and (hit["sstart"] == 1):
            return ("", "")
        # Query hanging off to the left
        if (hit["qstart"] > 1) and (hit["sstart"] == 1):
            endgap_len = hit["qstart"] - 1
            return (qseq[:endgap_len], "-" * endgap_len)
        # Subject hanging off to the left
        if (hit["qstart"] == 1) and (hit["sstart"] > 1):
            endgap_len = hit["sstart"] - 1
            return ("-" * endgap_len, sseq[:endgap_len])
        # Anything not meeting these conditions is bad
        if (hit["qstart"] > 1) and (hit["sstart"] > 1):
            raise ValueError("Unaligned sequence on left")
        raise ValueError("Query or subject start position less than 1")

    @staticmethod
    def _add_endgaps_right(hit, qseq, sseq):
        # No repair needed
        if (hit["qend"] == hit["qlen"]) and (hit["send"] == hit["slen"]):
            return ("", "")
        # Query hanging off to the right
        if (hit["qend"] < hit["qlen"]) and (hit["send"] == hit["slen"]):
            endgap_len = hit["qlen"] - hit["qend"]
            return (qseq[-endgap_len:], "-" * endgap_len)
        # Subject hanging off to the right
        if (hit["qend"] == hit["qlen"]) and (hit["send"] < hit["slen"]):
            endgap_len = hit["slen"] - hit["send"]
            return ("-" * endgap_len, sseq[-endgap_len:])
        # Anything not meeting these conditions is bad
        if (hit["qend"] < hit["qlen"]) and (hit["send"] < hit["qlen"]):
            raise ValueError("Unaligned sequence on right")
        raise ValueError("Query or subject end position greater than length")

    def _get_subject_seq(self, subject_id):
        subject_outfile = tempfile.NamedTemporaryFile()
        subject_outfile_fp = subject_outfile.name
        args = [
            "blastdbcmd",
            "-db",
            self.db,
            "-entry",
            subject_id,
            "-out",
            subject_outfile_fp,
        ]
        subprocess.check_call(args)
        with open(subject_outfile_fp) as f:
            return list(parse_fasta(f, trim_desc=True))[0][1]


def align_semiglobal(qseq, sseq):
    alignment = pairwise2.align.globalms(
        sseq,
        qseq,
        5,
        -4,
        -10,
        -0.5,  # match, mismatch, gapopen, gapextend
        penalize_end_gaps=False,
        one_alignment_only=True,
    )
    subj_seq = alignment[0][0]
    query_seq = alignment[0][1]
    return query_seq, subj_seq



import argparse
import os
import subprocess

import argparse
import collections
import itertools
import operator
import tempfile

from unassigner.parse import parse_fasta, write_fasta
from unassigner.align import VsearchAligner


class MismatchLocationApp:
    def __init__(
        self, species_file, ref_fp, mismatch_file, batch_size=10, num_cpus=None
    ):
        self.typestrain_seqs = list(parse_fasta(species_file, trim_desc=True))
        self.reference_fasta_fp = ref_fp
        self.mismatch_file = mismatch_file

        self.min_pct_id = 97.0
        self.num_threads = num_cpus
        self.batch_size = batch_size
        self.max_hits = "10000"

    @property
    def reference_udb_fp(self):
        base_fp, _ = os.path.splitext(self.reference_fasta_fp)
        return base_fp + ".udb"

    def make_reference_udb(self):
        if os.path.exists(self.reference_udb_fp):
            return None
        args = [
            "vsearch",
            "--makeudb_usearch",
            self.reference_fasta_fp,
            "--output",
            self.reference_udb_fp,
        ]
        return subprocess.check_call(args)

    def search_reference_seqs(self, query_seqs):
        query_file = tempfile.NamedTemporaryFile(suffix=".fasta", mode="wt")
        write_fasta(query_file, query_seqs)
        query_file.seek(0)

        reference_hits_file = tempfile.NamedTemporaryFile(suffix=".txt", mode="wt")

        # 97.0 --> 0.97
        vsearch_min_id = "{:.2f}".format(self.min_pct_id / 100)
        vsearch_args = [
            "vsearch",
            "--usearch_global",
            query_file.name,
            "--db",
            self.reference_udb_fp,
            "--userout",
            reference_hits_file.name,
            "--iddef",
            "2",
            "--id",
            vsearch_min_id,
            "--maxaccepts",
            self.max_hits,
            "--userfields",
            "query+target+id2+alnlen+mism+gaps+qilo+qihi+tilo+tihi+qs+ts+qrow+trow",
        ]
        if self.num_threads:
            vsearch_args.extend(["--threads", str(self.num_threads)])

        subprocess.check_call(vsearch_args)
        reference_hits_file.seek(0)
        return reference_hits_file

    def find_mismathes(self, reference_hits_file):
        ref_hits_file_read = open(reference_hits_file.name)
        hits = VsearchAligner._parse(ref_hits_file_read)
        for hit in hits:
            if hit["pident"] > 97.0:
                query_id = hit["qseqid"]
                subject_id = hit["sseqid"]
                mismatch_positions = mismatch_query_pos(hit)
                yield (query_id, subject_id, mismatch_positions)

    def run(self):
        self.make_reference_udb()
        for query_seqs in group_by_n(self.typestrain_seqs, self.batch_size):
            reference_hits_file = self.search_reference_seqs(query_seqs)
            ref_mismatches = self.find_mismathes(reference_hits_file)
            for query_id, subject_id, mismatch_positions in ref_mismatches:
                mismatch_positions = list(mismatch_positions)
                write_mismatches(
                    self.mismatch_file, query_id, subject_id, mismatch_positions
                )


def group_by_n(xs, n):
    args = [iter(xs)] * n
    for xgroup in itertools.zip_longest(*args, fillvalue=None):
        yield [x for x in xgroup if x is not None]


class MismatchDb(collections.abc.Mapping):
    def __init__(self, data=None):
        if data is None:
            self.data = {}
        else:
            self.data = data

    def __getitem__(self, key):
        if not key in self.data:
            self.data[key] = []
        return self.data[key]

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    @classmethod
    def load(cls, f):
        db = cls()
        for line in f:
            line = line.strip()
            if line.startswith("#"):
                continue
            if not line:
                continue
            vals = line.split("\t")
            query_id = vals[0]
            subject_id = vals[1]
            mismatch_positions = map(int, vals[2:])
            val = (subject_id, mismatch_positions)
            if query_id not in db.data:
                db.data[query_id] = [val]
            else:
                db.data[query_id].append(val)
        return db


def write_mismatches(f, query_id, subject_id, mismatch_positions):
    f.write(query_id)
    f.write("\t")
    f.write(subject_id)
    f.write("\t")
    f.write("\t".join(map(str, mismatch_positions)))
    f.write("\n")


class MutableMismatchDb(MismatchDb):
    def write(self, f):
        for query_id, mismatches in self.data.items():
            for subject_id, mismatch_positions in mismatches:
                f.write(query_id)
                f.write("\t")
                f.write(subject_id)
                f.write("\t")
                f.write("\t".join(map(str, mismatch_positions)))
                f.write("\n")

    def __setitem__(self, key, val):
        self.data[key] = val

    def __delitem__(self, key):
        del self.data[key]


AMBIGUOUS_BASES = {
    "-": "-",
    "T": "T",
    "C": "C",
    "A": "A",
    "G": "G",
    "R": "AG",
    "Y": "TC",
    "M": "CA",
    "K": "TG",
    "S": "CG",
    "W": "TA",
    "H": "TCA",
    "B": "TCG",
    "V": "CAG",
    "D": "TAG",
    "N": "TCAG",
}


def nucleotides_compatible(nt1, nt2):
    if (nt1 == "N") or (nt2 == "N"):
        return True
    a1 = AMBIGUOUS_BASES[nt1]
    a2 = AMBIGUOUS_BASES[nt2]
    return bool(set(a1).intersection(a2))


def hit_matches_by_alignment_pos(hit):
    qseq = hit["qseq"].upper()
    sseq = hit["sseq"].upper()
    query_pos = hit["qstart"]
    for q, s in zip(qseq, sseq):
        is_match = True
        if (q != s) and not nucleotides_compatible(q, s):
            is_match = False
        yield (query_pos, is_match)
        if q != "-":
            query_pos += 1


def hit_matches_by_query_pos(hit):
    query_pos_groups = itertools.groupby(
        hit_matches_by_alignment_pos(hit), operator.itemgetter(0)
    )
    for qpos, matches in query_pos_groups:
        all_matched = all(is_match for _, is_match in matches)
        yield qpos, all_matched


def mismatch_query_pos(hit):
    for qpos, is_match in hit_matches_by_query_pos(hit):
        if not is_match:
            yield qpos


def main(argv=None):
    p = argparse.ArgumentParser()
    p.add_argument(
        "type_strain_fasta",
        metavar="type-strain-fasta",
        type=argparse.FileType("r"),
        help="Type strain sequences FASTA file",
    )
    p.add_argument(
        "reference_fasta",
        metavar="reference-fasta",
        help="Full-length reference sequences FASTA file",
    )
    p.add_argument(
        "output_file",
        metavar="output-file",
        type=argparse.FileType("w"),
        help="Otuput file path",
    )

    p.add_argument(
        "--batch-size",
        type=int,
        default=10,
        help=(
            "Number of query sequences to search simultaneously "
            "(default: %(default)s)"
        ),
    )
    p.add_argument(
        "--num-cpus",
        type=int,
        help="Number of CPUs to use in search (default: all the CPUs)",
    )
    args = p.parse_args(argv)

    app = MismatchLocationApp(
        args.type_strain_fasta,
        args.reference_fasta,
        args.output_file,
        batch_size=args.batch_size,
        num_cpus=args.num_cpus,
    )
    app.run()
    args.output_file.close()



import logging
from io import StringIO


def parse_species_names(f):
    for desc, _ in parse_fasta(f):
        vals = desc.split("\t", maxsplit=1)
        accession = vals[0]
        if len(vals) == 2:
            species_name = vals[1]
        else:
            species_name = accession
        yield accession, species_name


def parse_fasta(f, trim_desc=False):
    """Parse a FASTA format file.

    Parameters
    ----------
    f : File object or iterator returning lines in FASTA format.

    Returns
    -------
    An iterator of tuples containing two strings
        First string is the sequence description, second is the
        sequence.

    Notes
    -----
    This function removes whitespace in the sequence and translates
    "U" to "T", in order to accommodate FASTA files downloaded from
    SILVA and the Living Tree Project.
    """
    f = iter(f)
    try:
        desc = next(f).strip()[1:]
        if trim_desc:
            desc = desc.split()[0]
    except StopIteration:
        return
    seq = StringIO()
    for line in f:
        line = line.strip()
        if line.startswith(">"):
            yield desc, seq.getvalue()
            desc = line[1:]
            if trim_desc:
                desc = desc.split()[0]
            seq = StringIO()
        else:
            seq.write(line.replace(" ", "").replace("U", "T"))
    yield desc, seq.getvalue()


def parse_desc(desc):
    try:
        arr = desc.split("|")
        accession = arr[2]
        species_name = arr[3]
        # This is the old way of parsing the description, 01_2022
        # accession = re.findall(r"\[accession=(.*?)\]", desc)[0]
        # species_name = re.findall(r"\[organism=(.*?)\]", desc)[0]
    except IndexError:
        logging.error(f"Couldn't find accession and/or organism identifier in {desc}")
        logging.error(f"Skipping this sequence...")
        return None, None
    return accession, species_name


def write_fasta(f, seqs):
    for desc, seq in seqs:
        f.write(">{0}\n{1}\n".format(desc, seq))


def load_fasta(filepath, trim_desc=True):
    """Load all sequences from a FASTA file

    Parameters
    ----------
    fasta_fp : Input filepath, FASTA format.

    Returns
    -------
    A dictionary mapping sequence identifiers to sequences.
    """
    with open(filepath) as f:
        seqs = parse_fasta(f, trim_desc=trim_desc)
        return dict(seqs)


def parse_greengenes_accessions(f):
    for line in f:
        if line.startswith("#"):
            continue
        line = line.strip()
        yield line.split("\t")


def cast_num_or_na(val, cast_func):
    if val == "NA":
        return None
    return cast_func(val)


def parse_results(f):
    float_fields = ["probability_incompatible"]
    int_fields = ["region_mismatches", "region_positions"]
    header_line = next(f)
    header_line = header_line.rstrip()
    fields = header_line.split("\t")
    for line in f:
        line = line.rstrip()
        vals = line.split("\t")
        res = dict(zip(fields, vals))
        for field, val in res.items():
            if field in float_fields:
                res[field] = cast_num_or_na(val, float)
            elif field in int_fields:
                res[field] = cast_num_or_na(val, int)
        yield res



import argparse
import sys
import os

from unassigner.download import (
    get_url,
    clean,
    LTP_METADATA_URL,
    LTP_SEQS_URL,
    GG_SEQS_URL,
    GG_ACCESSIONS_URL,
    process_ltp_seqs,
    process_greengenes_seqs,
)


def use_or_download(optional_fp, url, db_dir):
    if optional_fp is None:
        return get_url(url, os.path.join(db_dir, os.path.basename(url)))
    else:
        return optional_fp


def main(argv=None):
    p = argparse.ArgumentParser()
    p.add_argument(
        "--ltp_metadata_fp",
        help=(
            "Filepath for LTP metadata (.csv file) "
            "[default:download from LTP website]"
        ),
    )
    p.add_argument(
        "--ltp_seqs_fp",
        help=(
            "Filepath for unaligned 16S sequences from LTP (.fasta file) "
            "[default: download from LTP website]"
        ),
    )
    p.add_argument(
        "--download_greengenes",
        action="store_true",
        help="Download GreenGenes reference files [default: False]",
    )
    p.add_argument(
        "--greengenes_accessions_fp",
        help=(
            "Filepath for table of GreenGenes accession numbers "
            "(.txt or .txt.gz file) "
            "[default: download from GreenGenes mirror]"
        ),
    )
    p.add_argument(
        "--greengenes_seqs_fp",
        help=(
            "Filepath for unaligned 16S reference sequences "
            "(.fasta or .fasta.gz file) "
            "[default: download from GreenGenes mirror]"
        ),
    )
    p.add_argument(
        "--clean",
        action="store_true",
        help=("Remove all downloaded and processed files."),
    )
    p.add_argument("--db-dir", help=("Filepath to download the files to."))
    args = p.parse_args(argv)

    if args.db_dir:
        db_dir = args.db_dir
        if not os.path.exists(args.db_dir) and not args.clean:
            os.mkdir(args.db_dir)
    else:
        db_dir = os.getcwd()

    if args.clean is True:
        clean(db_dir)
        sys.exit(0)

    use_or_download(args.ltp_metadata_fp, LTP_METADATA_URL, db_dir)
    ltp_seqs_fp = use_or_download(args.ltp_seqs_fp, LTP_SEQS_URL, db_dir)
    process_ltp_seqs(ltp_seqs_fp, db_dir)

    if args.download_greengenes:
        gg_seqs_fp = use_or_download(args.greengenes_seqs_fp, GG_SEQS_URL, db_dir)
        gg_accessions_fp = use_or_download(
            args.greengenes_accessions_fp, GG_ACCESSIONS_URL, db_dir
        )
        process_greengenes_seqs(gg_seqs_fp, gg_accessions_fp, db_dir)


def download_type_strain_data(output_dir=None, metadata_fp=None, seqs_fp=None):
    if output_dir is None:
        output_dir = os.getcwd()
    metadata_fp = use_or_download(metadata_fp, LTP_METADATA_URL, output_dir)
    seqs_fp = use_or_download(seqs_fp, LTP_SEQS_URL, output_dir)
    return process_ltp_seqs(seqs_fp, output_dir)



import argparse
import sys

from unassigner.parse import parse_fasta, write_fasta
from unassigner.trim import (
    CompleteMatcher,
    PartialMatcher,
    deambiguate,
    reverse_complement,
)


def main(argv=None):
    p = argparse.ArgumentParser()
    p.add_argument("--input_file", type=argparse.FileType("r"), default=sys.stdin)
    p.add_argument("--output_file", type=argparse.FileType("w"), default=sys.stdout)
    p.add_argument(
        "--unmatched_output_file",
        type=argparse.FileType("w"),
    )
    p.add_argument("--query", required=True)
    p.add_argument(
        "--max_mismatch",
        type=int,
        default=2,
        help="Maximum number of mismatches in complete match",
    )
    p.add_argument(
        "--min_partial",
        type=int,
        default=5,
        help="Minimum length of partial sequence match",
    )
    p.add_argument(
        "--reverse_complement_query",
        action="store_true",
        help="Reverse complement the query seq before search",
    )

    args = p.parse_args(argv)

    queryset = deambiguate(args.query)
    if args.reverse_complement_query:
        queryset = [reverse_complement(q) for q in queryset]

    seqs = parse_fasta(args.input_file)

    cm = CompleteMatcher(queryset, args.max_mismatch)
    unmatched_cm = []
    for seq_id, seq in seqs:
        m = cm.find_match(seq)
        if m is None:
            unmatched_cm.append((seq_id, seq))
        else:
            write_fasta(args.output_file, [(seq_id, seq)])

    if args.min_partial > 0:
        pm = PartialMatcher(queryset, args.min_partial)
        unmatched_pm = []
        for seq_id, seq in unmatched_cm:
            m = pm.find_match(seq)
            if m is None:
                unmatched_pm.append((seq_id, seq))
            else:
                write_fasta(args.output_file, [(seq_id, seq)])
    else:
        unmatched_pm = unmatched_cm

    if args.unmatched_output_file is not None:
        write_fasta(args.unmatched_output_file, unmatched_pm)



import itertools


class AlignedPair(object):
    def __init__(self, qseq, sseq):
        self.query_id, self.query_seq = qseq
        self.subject_id, self.subject_seq = sseq
        assert len(self.query_seq) == len(self.subject_seq)

    def __eq__(self, other):
        self_vals = (self.query_id, self.query_seq, self.subject_id, self.subject_seq)
        other_vals = (
            other.query_id,
            other.query_seq,
            other.subject_id,
            other.subject_seq,
        )
        return self_vals == other_vals

    @property
    def alignment_len(self):
        return len(self.query_seq)

    @property
    def query_len(self):
        return len(self.query_seq) - self.query_seq.count("-")

    @property
    def subject_len(self):
        return len(self.subject_seq) - self.subject_seq.count("-")

    @property
    def unaligned_query_seq(self):
        return self.query_seq.replace("-", "")

    @property
    def unaligned_subject_seq(self):
        return self.subject_seq.replace("-", "")

    def count_matches(self):
        return sum(q == s for q, s in zip(self.query_seq, self.subject_seq))

    @property
    def percent_id(self):
        return self.count_matches() / self.query_len

    def trim_endgaps(self):
        return AlignedRegion.without_endgaps(self).trim_ends()


class AlignedRegion:
    """Specify and extract regions of a pairwise sequence alignment"""

    def __init__(self, alignment, start_idx, end_idx):
        assert start_idx >= 0
        assert start_idx <= alignment.alignment_len
        assert end_idx >= 0
        assert end_idx <= alignment.alignment_len
        assert start_idx <= end_idx
        self.alignment = alignment
        self.start_idx = start_idx
        self.end_idx = end_idx

    def trim_ends(self):
        qseq = self.alignment.query_seq[self.start_idx : self.end_idx]
        sseq = self.alignment.subject_seq[self.start_idx : self.end_idx]
        return AlignedPair(
            (self.alignment.query_id, qseq), (self.alignment.subject_id, sseq)
        )

    def trim_left(self, include_region=False):
        if include_region:
            idx = self.start_idx
        else:
            idx = self.end_idx
        qseq = self.alignment.query_seq[idx:]
        sseq = self.alignment.subject_seq[idx:]
        return AlignedPair(
            (self.alignment.query_id, qseq), (self.alignment.subject_id, sseq)
        )

    def trim_right(self, include_region=False):
        if include_region:
            idx = self.end_idx
        else:
            idx = self.start_idx
        qseq = self.alignment.query_seq[:idx]
        sseq = self.alignment.subject_seq[:idx]
        return AlignedPair(
            (self.alignment.query_id, qseq), (self.alignment.subject_id, sseq)
        )

    def in_alignment(self):
        return (self.start_idx, self.end_idx)

    def in_subject(self):
        return in_aligned_seq(self.alignment.subject_seq, self.start_idx, self.end_idx)

    def in_query(self):
        return in_aligned_seq(self.alignment.query_seq, self.start_idx, self.end_idx)

    def subject_offset(self):
        qstart, qend = self.in_subject()
        qlen = self.alignment.subject_len
        if (qstart == 0) and (qend == 0):
            # Subject is off to right side, offset is positive
            subject_right = self.alignment.subject_seq[self.end_idx :]
            return count_endgaps(subject_right)
        if (qstart == qlen) and (qend == qlen):
            # Subject is off to left side, offset is negative
            subject_left = self.alignment.subject_seq[: self.start_idx]
            return -count_endgaps(subject_left, reverse=True)
        return 0

    def query_offset(self):
        qstart, qend = self.in_query()
        qlen = self.alignment.query_len
        if (qstart == 0) and (qend == 0):
            # Query is off to right side, offset is positive
            query_right = self.alignment.query_seq[self.end_idx :]
            return count_endgaps(query_right)
        if (qstart == qlen) and (qend == qlen):
            # Query is off to left side, offset is negative
            query_left = self.alignment.query_seq[: self.start_idx]
            return -count_endgaps(query_left, reverse=True)
        return 0

    @classmethod
    def without_endgaps(cls, a):
        """Create a new region with no endgaps"""
        left_endgaps = max(count_endgaps(a.query_seq), count_endgaps(a.subject_seq))
        right_endgaps = max(
            count_endgaps(a.query_seq, reverse=True),
            count_endgaps(a.subject_seq, reverse=True),
        )
        start_idx = left_endgaps
        end_idx = a.alignment_len - right_endgaps
        return cls(a, start_idx, end_idx)

    @classmethod
    def from_subject(cls, a, subject_start_idx=0, subject_end_idx=None):
        """Create a new region using coordinates in subject sequence"""
        if subject_end_idx is None:
            subject_end_idx = a.subject_len
        assert subject_start_idx >= 0
        assert subject_start_idx <= a.subject_len
        assert subject_end_idx >= 0
        assert subject_end_idx <= a.subject_len
        assert subject_start_idx <= subject_end_idx
        start_idx = aligned_start_idx(a.subject_seq, subject_start_idx)
        end_idx = aligned_end_idx(a.subject_seq, subject_end_idx)
        return cls(a, start_idx, end_idx)

    @classmethod
    def from_query(cls, a, query_start_idx=0, query_end_idx=None):
        """Create a new region using coordinates in query sequence"""
        if query_end_idx is None:
            query_end_idx = a.query_len
        assert query_start_idx >= 0
        assert query_start_idx <= a.query_len
        assert query_end_idx >= 0
        assert query_end_idx <= a.query_len
        assert query_start_idx <= query_end_idx
        start_idx = aligned_start_idx(a.query_seq, query_start_idx)
        end_idx = aligned_end_idx(a.query_seq, query_end_idx)
        return cls(a, start_idx, end_idx)


def in_aligned_seq(seq, start_idx, end_idx):
    seq_start = seq[:start_idx]
    seq_start_idx = start_idx - seq_start.count("-")
    seq_end = seq[:end_idx]
    seq_end_idx = end_idx - seq_end.count("-")
    return seq_start_idx, seq_end_idx


def aligned_end_idx(seq, end_idx):
    # Work in reverse mode, find the complimentary start_idx
    ungapped_seq_len = len(seq) - seq.count("-")
    rev_start_idx = ungapped_seq_len - end_idx
    rev_seq = seq[::-1]
    rev_alignment_start_idx = aligned_start_idx(rev_seq, rev_start_idx)
    return len(seq) - rev_alignment_start_idx


def aligned_start_idx(seq, start_idx):
    ungapped_seq_len = len(seq) - seq.count("-")
    # Special case: start_idx is equal to ungapped sequence length
    # We will never count enough bases to reach the start_idx
    # Reverse the sequence and find the start position for index 0,
    # then subtract from the alignment length to get the forward-facing
    # index after the last base.
    if start_idx == ungapped_seq_len:
        rev_seq = seq[::-1]
        rev_alignment_start_idx = aligned_start_idx(rev_seq, 0)
        return len(seq) - rev_alignment_start_idx
    start_idxs = [n - 1 for n in cumulative_bases(seq)]
    return start_idxs.index(start_idx)


def cumulative_bases(seq):
    return itertools.accumulate(int(b != "-") for b in seq)


def iter_len(xs):
    n = 0
    for _ in xs:
        n += 1
    return n


def count_endgaps(seq, reverse=False):
    if reverse:
        seq = reversed(seq)
    return iter_len(itertools.takewhile(lambda x: x == "-", seq))



import argparse
import collections
import os
import os.path
import random
import re
import shutil
import subprocess
import urllib.error

from unassigner.download import get_url
from unassigner.parse import parse_fasta, write_fasta


class Refseq16SDatabase:
    def __init__(
        self, fasta_fp="refseq_16S.fasta", accession_fp="refseq_16S_accessions.txt"
    ):
        self.fasta_fp = fasta_fp
        self.accession_fp = accession_fp
        self.seqs = {}
        self.assemblies = {}
        self.seqids_by_assembly = collections.defaultdict(list)

    def add_assembly(self, assembly, select_random=False):
        seqs = list(assembly.ssu_seqs)
        if select_random and (len(seqs) > 0):
            seqs = [random.choice(seqs)]

        # Avoid writing duplicate genes for the same genome
        seen = set()
        for desc, seq in seqs:
            if seq not in seen:
                print(desc)
                seqid = desc.split()[0]
                self.assemblies[seqid] = assembly
                self.seqs[seqid] = seq
                self.seqids_by_assembly[assembly.accession].append(seqid)
                seen.add(seq)

    def load(self, assemblies):
        with open(self.accession_fp, "r") as f:
            for line in f:
                toks = line.strip().split()
                seqid = toks[0]
                accession = toks[1]
                assembly = assemblies[accession]
                self.assemblies[seqid] = assembly
                self.seqids_by_assembly[assembly.accession].append(seqid)
        with open(self.fasta_fp, "r") as f:
            for seqid, seq in parse_fasta(f):
                self.seqs[seqid] = seq

    def save(self):
        with open(self.fasta_fp, "w") as f:
            write_fasta(f, self.seqs.items())
        with open(self.accession_fp, "w") as f:
            for seqid, assembly in self.assemblies.items():
                f.write("{0}\t{1}\n".format(seqid, assembly.accession))

    def compute_pctids(self, min_pctid=97.0, threads=None):
        aligner = PctidAligner(self.fasta_fp)
        if not os.path.exists(aligner.hits_fp):
            aligner.search(min_pctid=min_pctid, threads=threads)
        with open(aligner.hits_fp) as f:
            for hit in aligner.parse(f):
                query = self.assemblies[hit["qseqid"]]
                subject = self.assemblies[hit["sseqid"]]
                pctid = hit["pident"]
                yield AssemblyPair(query, subject, pctid)

    def search_one(self, query_seqid, pctid, threads=None):
        pctid_str = "{:.1f}".format(pctid)
        print("Searching", query_seqid, "at", pctid_str, "pct identity")
        query_seq = self.seqs[query_seqid]
        query_fp = "temp_query.fasta"
        if os.path.exists(query_fp):
            os.rename(query_fp, "temp_prev_query.fasta")
        query_hits_fp = "temp_query_hits.txt"
        if os.path.exists(query_hits_fp):
            os.rename(query_hits_fp, "temp_prev_query_hits.txt")
        with open(query_fp, "w") as f:
            write_fasta(f, [(query_seqid, query_seq)])
        aligner = PctidAligner(self.fasta_fp)
        aligner.search(
            query_fp, query_hits_fp, min_pctid=pctid, threads=threads, max_hits=10000
        )
        with open(query_hits_fp) as f:
            hits = aligner.parse(f)
            for hit in hits:
                if hit["pident"] == pctid_str:
                    query = self.assemblies[hit["qseqid"]]
                    subject = self.assemblies[hit["sseqid"]]
                    pctid = hit["pident"]
                    yield AssemblyPair(
                        query, subject, pctid, hit["qseqid"], hit["sseqid"]
                    )

    def search_seq(self, query_seqid, query_seq, min_pctid=90.0, threads=None):
        query_fp = "temp_query.fasta"
        if os.path.exists(query_fp):
            os.rename(query_fp, "temp_prev_query.fasta")
        query_hits_fp = "temp_query_hits.txt"
        if os.path.exists(query_hits_fp):
            os.rename(query_hits_fp, "temp_prev_query_hits.txt")
        with open(query_fp, "w") as f:
            write_fasta(f, [(query_seqid, query_seq)])
        aligner = PctidAligner(self.fasta_fp)
        aligner.search(
            query_fp,
            query_hits_fp,
            min_pctid=min_pctid,
            threads=threads,
            max_hits=10000,
        )
        with open(query_hits_fp) as f:
            hits = aligner.parse(f)
            for hit in hits:
                query = self.assemblies[hit["qseqid"]]
                subject = self.assemblies[hit["sseqid"]]
                pctid = hit["pident"]
                if query.accession != subject.accession:
                    yield AssemblyPair(
                        query, subject, pctid, hit["qseqid"], hit["sseqid"]
                    )


class PctidAligner:
    field_names = ["qseqid", "sseqid", "pident"]
    hits_fp = "refseq_16S_hits.txt"

    def __init__(self, fasta_fp):
        self.fasta_fp = fasta_fp

    @property
    def reference_udb_fp(self):
        base_fp, _ = os.path.splitext(self.fasta_fp)
        return base_fp + ".udb"

    def make_reference_udb(self):
        if os.path.exists(self.reference_udb_fp):
            return None
        args = [
            "vsearch",
            "--makeudb_usearch",
            self.fasta_fp,
            "--output",
            self.reference_udb_fp,
        ]
        return subprocess.check_call(args)

    def search(
        self, input_fp=None, hits_fp=None, min_pctid=97.0, threads=None, max_hits=10000
    ):
        if input_fp is None:
            input_fp = self.fasta_fp
        if hits_fp is None:
            hits_fp = self.hits_fp
        self.make_reference_udb()
        # 97.0 --> 0.97
        min_id = "{:.3f}".format(min_pctid / 100)
        args = [
            "vsearch",
            "--usearch_global",
            input_fp,
            "--db",
            self.reference_udb_fp,
            "--userout",
            hits_fp,
            "--iddef",
            "2",
            "--id",
            min_id,
            "--userfields",
            "query+target+id2",
        ]
        if max_hits is None:
            args.extend(
                [
                    "--maxaccepts",
                    "0",
                    "--maxrejects",
                    "0",
                ]
            )
        else:
            args.extend(
                [
                    "--maxaccepts",
                    str(max_hits),
                ]
            )
        if threads is not None:
            args.extend(["--threads", str(threads)])
        print(args)
        subprocess.check_call(args)
        return hits_fp

    def parse(self, f):
        for line in f:
            line = line.strip()
            if line.startswith("#"):
                continue
            vals = line.split("\t")
            hit = dict(zip(self.field_names, vals))
            if hit["qseqid"] != hit["sseqid"]:
                yield hit


class RefseqAssembly:
    summary_url = (
        "https://ftp.ncbi.nlm.nih.gov/genomes/refseq/bacteria/assembly_summary.txt"
    )
    summary_fp = "refseq_bacteria_assembly_summary.txt"
    genome_dir = "genome_fasta"
    rna_dir = "rna_fasta"
    summary_cols = [
        "assembly_accession",
        "bioproject",
        "biosample",
        "wgs_master",
        "refseq_category",
        "taxid",
        "species_taxid",
        "organism_name",
        "infraspecific_name",
        "isolate",
        "version_status",
        "assembly_level",
        "release_type",
        "genome_rep",
        "seq_rel_date",
        "asm_name",
        "submitter",
        "gbrs_paired_asm",
        "paired_asm_comp",
        "ftp_path",
        "excluded_from_refseq",
        "relation_to_type_material",
    ]

    def __init__(self, assembly_accession, ftp_path, **kwargs):
        self.accession = assembly_accession
        self.ftp_path = ftp_path
        for key, val in kwargs.items():
            setattr(self, key, val)
        self._ssu_seqs = None

    @classmethod
    def parse_summary(cls, f):
        for line in f:
            line = line.rstrip("\n")
            if line.startswith("#") or (line == ""):
                continue
            toks = line.split("\t")
            vals = dict(zip(cls.summary_cols, toks))
            if vals["ftp_path"] == "na":
                continue
            yield cls(**vals)

    @property
    def ssu_seqs(self):
        if self._ssu_seqs is not None:
            return self._ssu_seqs
        if not os.path.exists(self.rna_fp):
            try:
                self.download_rna()
            except urllib.error.HTTPError as e:
                print(self.accession)
                print(e)
                return []
        with open(self.rna_fp, "rt") as f:
            seqs = list(parse_fasta(f))
        res = [(desc, seq) for (desc, seq) in seqs if is_16S(desc)]
        self._ssu_seqs = res
        return res

    @classmethod
    def load(cls):
        if not os.path.exists(cls.summary_fp):
            get_url(cls.summary_url, cls.summary_fp)
        with open(cls.summary_fp, "r") as f:
            return {a.accession: a for a in cls.parse_summary(f)}

    @property
    def base_url(self):
        return re.sub("^ftp://", "https://", self.ftp_path)

    @property
    def basename(self):
        return os.path.basename(self.ftp_path)

    @property
    def rna_url(self):
        return "{0}/{1}_rna_from_genomic.fna.gz".format(self.base_url, self.basename)

    @property
    def rna_fp(self):
        rna_filename = "{0}_rna_from_genomic.fna".format(self.basename)
        return os.path.join(self.rna_dir, rna_filename)

    def download_rna(self):
        if os.path.exists(self.rna_fp):
            return
        if not os.path.exists(self.rna_dir):
            os.mkdir(self.rna_dir)
        print("Downloading 16S seqs for ", self.accession)
        get_url(self.rna_url, self.rna_fp + ".gz")
        subprocess.check_call(["gunzip", "-q", self.rna_fp + ".gz"])

    @property
    def genome_url(self):
        return "{0}/{1}_genomic.fna.gz".format(self.base_url, self.basename)

    @property
    def genome_fp(self):
        genome_filename = "{0}_genomic.fna.gz".format(self.basename)
        return os.path.join(self.genome_dir, genome_filename)

    def download_genome(self, genome_dir=None, genome_fname=None):
        if genome_dir is None:
            genome_dir = self.genome_dir
        if genome_fname is None:
            genome_fp = self.genome_fp
        else:
            genome_fp = os.path.join(genome_dir, genome_fname)
        if os.path.exists(genome_fp):
            return
        if not os.path.exists(genome_dir):
            os.mkdir(genome_dir)
        get_url(self.genome_url, genome_fp)


def is_16S(desc):
    return "product=16S ribosomal RNA" in desc


def subsample_by(xs, fcn, n):
    groups = group_by(xs, fcn)
    for group in groups:
        if len(group) > n:
            group = random.sample(group, n)
        for x in group:
            yield x


def flatten(xss):
    return [x for xs in xss for x in xs]


def group_by(xs, fcn):
    groups = collections.defaultdict(list)
    for x in xs:
        group = fcn(x)
        groups[group].append(x)
    return groups.values()


def remove_files(target_dir):
    if os.path.exists(target_dir):
        for filename in os.listdir(target_dir):
            os.remove(os.path.join(target_dir, filename))


class AssemblyPair:
    genome_dir = "temp_genomes"
    ani_dir = "temp_ani"
    cache_dir = "temp_genomes_cache"
    ani_cache = {}

    def __init__(
        self, query, subject, pctid=None, query_seqid=None, subject_seqid=None
    ):
        self.query = query
        self.subject = subject
        self._pctid = pctid
        self.ani = None
        self.query_seqid = query_seqid
        self.subject_seqid = subject_seqid

    @property
    def pctid(self):
        return round(float(self._pctid), 1)

    def compute_ani(self):
        ani_key = (self.query.accession, self.subject.accession)
        if ani_key in self.ani_cache:
            print("ANI is cached")
            self.ani = self.ani_cache[ani_key]
            return

        remove_files(self.genome_dir)

        query_fname = "{0}.fna.gz".format(self.query.accession)
        query_genome_fp = os.path.join(self.genome_dir, query_fname)
        query_cache_fp = os.path.join(self.cache_dir, query_fname)
        if os.path.exists(query_cache_fp):
            os.rename(query_cache_fp, query_genome_fp)
        else:
            self.query.download_genome(self.genome_dir, query_fname)
            shutil.copy(query_genome_fp, query_cache_fp)
        subprocess.check_call(["gunzip", "-f", query_genome_fp])

        subject_fname = "{0}.fna.gz".format(self.subject.accession)
        subject_genome_fp = os.path.join(self.genome_dir, subject_fname)
        self.subject.download_genome(self.genome_dir, subject_fname)
        subprocess.check_call(["gunzip", "-f", subject_genome_fp])

        # pyani get ANI for pair
        shutil.rmtree(self.ani_dir)
        # save ANI to self.ani
        print("Computing ANI for", self.query.accession, "and", self.subject.accession)
        subprocess.check_call(
            [
                "average_nucleotide_identity.py",
                "-i",
                self.genome_dir,
                "-o",
                self.ani_dir,
            ]
        )

        ani_fp = os.path.join(self.ani_dir, "ANIm_percentage_identity.tab")
        with open(ani_fp) as f:
            ani = parse_pairwise_ani(f)
        print("ANI:", ani)
        self.ani_cache[ani_key] = ani
        self.ani = ani

    def format_output(self):
        if self.query_seqid and self.subject_seqid:
            return "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\n".format(
                self.query.accession,
                self.subject.accession,
                self.query_seqid,
                self.subject_seqid,
                self.pctid,
                self.ani,
            )
        return "{0}\t{1}\t{2}\t{3}\n".format(
            self.query.accession,
            self.subject.accession,
            self.pctid,
            self.ani,
        )


def parse_pairwise_ani(f):
    # first line is a header
    next(f)
    # second line has ani as the third token
    second_line = next(f)
    toks = second_line.strip().split()
    return toks[2]


def pctid_range(min_pctid):
    assert 50.0 < min_pctid <= 100.0
    current_val = 100.0
    while current_val > min_pctid:
        yield current_val
        current_val = current_val - 0.1


def main_sampling(argv=None):
    p = argparse.ArgumentParser()
    p.add_argument(
        "--output-file",
        type=argparse.FileType("w"),
        default="refseq_pctid_ani.tsv",
        help="Output file",
    )
    p.add_argument(
        "--min_pctid",
        type=float,
        default=97.0,
        help="Minimum 16S percent ID",
    )
    p.add_argument(
        "--num-threads",
        type=int,
        help="Number of threads for 16S percent ID (default: use all CPUs)",
    )
    p.add_argument(
        "--num-ani",
        type=int,
        default=100,
        help="Number of genome pairs on which to evaluate ANI",
    )
    p.add_argument(
        "--seed",
        type=int,
        default=42,
        help="Random number seed",
    )
    args = p.parse_args()
    args.output_file.write(
        "query_assembly\tsubject_assembly\t"
        "query_seqid\tsubject_seqid\t"
        "pctid\tani\n"
    )

    # Set seed for 16S selection
    random.seed(args.seed)

    # Load all assemblies
    assemblies = RefseqAssembly.load()

    # 16S database with one random sequence from each assembly
    db = Refseq16SDatabase("refseq_16S_all.fasta", "refseq_16S_accessions_all.txt")
    if os.path.exists(db.accession_fp):
        db.load(assemblies)
    else:
        for assembly in assemblies.values():
            db.add_assembly(assembly)
        db.save()

    pctid_vals = list(pctid_range(args.min_pctid)) * args.num_ani

    # Set seed again
    random.seed(args.seed + 1)
    assembly_list = list(assemblies.values())
    for current_pctid in pctid_vals:
        found = False
        while not found:
            # randomly select assembly
            query_assembly = random.choice(assembly_list)
            # randomly select query sequence from assembly
            query_assembly_seqids = db.seqids_by_assembly[query_assembly.accession]
            # next loop if we don't have any sequences for this assembly
            if not query_assembly_seqids:
                continue
            query_seqid = random.choice(query_assembly_seqids)
            assembly_pairs = db.search_one(
                query_seqid, current_pctid, threads=args.num_threads
            )
            assembly_pairs = list(assembly_pairs)
            if assembly_pairs:
                try:
                    # randomly select one result
                    selected_pair = random.choice(assembly_pairs)
                    selected_pair.compute_ani()
                    args.output_file.write(selected_pair.format_output())
                    args.output_file.flush()
                except Exception as e:
                    print(e)
                else:
                    found = True



import os
from setuptools import setup

PACKAGE_DIR = os.path.abspath(os.path.dirname(__file__))
README_FP = os.path.join(PACKAGE_DIR, "README.md")
with open(README_FP, encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="unassigner",
    version="1.0.2",
    description="Bacterial species unassigner",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Kyle Bittinger",
    author_email="kylebittinger@gmail.com",
    url="https://github.com/kylebittinger/unassigner",
    packages=["unassigner"],
    install_requires=[
        "biopython",
        "scipy",
    ],
    entry_points={
        "console_scripts": [
            "unassign=unassigner.command:main",
            "trimragged=unassigner.trim:main",
            "count_mismatches=unassigner.mismatch_db:main",
            "pctid_ani_sample=unassigner.ani:main_sampling",
        ],
    },
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    license="GPLv2+",
)
